# A character sprite
from __future__ import division

import data
from cocos import sprite
from cocos.actions import Action, IntervalAction, MoveTo
from cocos.collision_model import AARectShape, CircleShape
from cocos.euclid import Vector2
from cocos.rect import Rect
from pyglet import image
import pyglet
from string import lower
import copy

# Image broken down into the following areas
# Stand    Look Up      Walk 1       Walk 2
# Talk     Look Up/Talk Blink        Blink/Talk
# Stand    Stand        Walk 1/Blink Walk 2/Blink
# Stand    Stand        Stand        Stand
_STAND = 0
_LOOK_UP = 1
_WALK_1 = 2
_WALK_2 = 3
_TALK = 4
_LOOK_UP_TALK = 5
_BLINK = 6
_BLINK_TALK = 7
_STAND_NORTH = 8
_WALK_1_BLINK = 10
_WALK_2_BLINK = 11
_WALK_1_NORTH = 12
_WALK_2_NORTH = 13
_REGION = ( # Is X, Y, Width, Height (allows for get_region(*_REGION[x]))
    (0, 192, 32, 64), (32, 192, 32, 64), (64, 192, 32, 64), (96, 192, 32, 64),
    (0, 128, 32, 64), (32, 128, 32, 64), (64, 128, 32, 64), (96, 128, 32, 64),
    (0, 64, 32, 64), (32, 64, 32, 64), (64, 64, 32, 64), (96, 64, 32, 64),
    (0, 0, 32, 64), (32,0, 32, 64), (64,0, 32, 64), (96, 0, 32, 64),
    )
_FRAME_COUNT = 16
back_set = ('north', 'northeast', 'northwest')

class WazSpawn(Action):
    """implements spawn when the result cannot be expresed as IntervalAction
        At least one operand must have duration==None """
    def init(self, one, two):
        one = copy.deepcopy(one)
        two = copy.deepcopy(two)
        self.actions = [one, two]

    def start(self):
        for action in self.actions:
            action.target = self.target
            action.start()

    def step(self, dt):
        for a in self.actions:
            if a.done():
                self._done = True
                self.stop()
                return
        
        for a in self.actions:
            a.step(dt)
        
    def stop(self):
        for e in self.actions:
            #print 'en el loop, name:', e.name
            e.stop()

class Blink(IntervalAction):
    def init(self):
        self.duration = 0.5
    
    def start(self):
        self._initial = self.target.frame
        if self._initial == _WALK_1: # walking
            self.target.frame = _WALK_1_BLINK
        elif self._initial == _WALK_2: # walking
            self.target.frame = _WALK_2_BLINK
        elif self._initial == (_TALK or _LOOK_UP_TALK):
            self.target.frame = _BLINK_TALK
        else:
            self.target.frame = _BLINK
    
    def stop(self):
        if self.target.frame == _WALK_1_BLINK:
            self.target.frame = _WALK_2
        elif self.target.frame == _WALK_2_BLINK:
            self.target.frame = _WALK_1
        else:
            self.target.frame = self._initial
    
    def update(self, t):
        pass
        
    def __reversed__(self):
        return Blink()

class March(Action):
    def init(self):
        pass

    def start(self):
        self._initial = True

    def step(self, dt):
        if self._initial or self.target.blocked:
            self._initial = False
            return
        # Move
        x, y = self.target.position
        dx, dy = self.target.velocity
        x += dx * dt
        y += dy * dt
        self.target.position = x, y

class Step(Action):
    def init(self):
        self._switch_time = 0.5
        self._elapsed = 0.0
        
    def start(self):
        self._initial = True
        if self.target.direction in back_set:
            print "WALK: Moving away from the camera"
            self.target.frame = _WALK_1_NORTH
        elif self.target.frame == _BLINK:
            print "WALK: Was blinking, so now walking and blinking"
            self.target.frame = _WALK_1_BLINK
        else:
            print "WALK: Now walking"
            self.target.frame = _WALK_1
    
    def step(self, dt):
        #animate
        self._elapsed += dt
        if self._elapsed > self._switch_time:
            self._elapsed = self._elapsed - self._switch_time
            if self.target.frame == _WALK_1_NORTH:
                self.target.frame = _WALK_2_NORTH
            elif self.target.frame == _WALK_2_NORTH:
                self.target.frame = _WALK_1_NORTH
            elif self.target.frame == _WALK_1:
                self.target.frame = _WALK_2
            elif self.target.frame == _WALK_1_BLINK:
                self.target.frame = _WALK_2_BLINK
            elif self.target.frame == _WALK_2_BLINK:
                self.target.frame = _WALK_1_BLINK
            else:
                self.target.frame = _WALK_1
    
    def stop(self):
        if self.target.frame == (_WALK_1_BLINK or _WALK_2_BLINK):
            self.target.frame = _BLINK
        else:
            self.target.frame = _STAND

def Walk():
    return WazSpawn(March(), Step())

def WalkTo(location, duration=5):
    return WazSpawn(Step(), MoveTo(location, duration))
            
class Direction(object):
    Stop = 0
    North = 2
    South = 3
    East  = 5
    West  = 7
    NorthEast = 10
    NorthWest = 14
    SouthEast = 15
    SouthWest = 21
    _dir_names = {
        'stop': Stop,
        'north': North,
        'south': South,
        'east': East,
        'west': West,
        'northeast': NorthEast,
        'northwest': NorthWest,
        'southeast': SouthEast,
        'southwest': SouthWest,
    }
    _speed = 96
    _velocity = {
        Stop: (0, 0),
        North: (0, _speed),
        South: (0, -_speed),
        East: (_speed, 0),
        West: (-_speed, 0),
        NorthEast: (_speed, _speed),
        SouthEast: (_speed, -_speed),
        NorthWest: (-_speed, _speed),
        SouthWest: (-_speed, -_speed),
    }
    def DirFromName(self, name):
        n = lower(name)
        return self._dir_names[n]
    
    def _less_first(self, dir1, dir2):
        return (dir1, dir2) if dir1 < dir2 else (dir2, dir1)
        
    def CanCombineDirs(self, dir1, dir2):
        d1, d2 = self._less_first(dir1, dir2)
        return (d1 in (self.North, self.South)) and (d2 in (self.East, self.West))
        
    def CombineDirs(self, dir1, dir2):
        if self.CanCombineDirs(dir1, dir2):
            return dir1 * dir2
        elif (dir1 == self.Stop) or (dir2 == self.Stop):
           return dir1 + dir2
        else:
            return dir1
            
    def SubtractDirs(self, dir1, dir2):
        if dir1 == self.Stop: return dir2
        if dir2 == self.Stop: return dir1
        if dir1 == dir2: return self.Stop
        d1, d2 = self._less_first(dir1, dir2)
        if ((d1 == self.North) and (d2 in (self.NorthEast, self.NorthWest))) \
        or ((d1 == self.South) and (d2 in (self.SouthEast, self.SouthWest))) \
        or ((d1 == self.East) and (d2 in (self.NorthEast, self.SouthEast))) \
        or ((d1 == self.West) and (d2 in (self.NorthWest, self.SouthWest))):
            return int(d2/d1)
        else:
            return d2
        
    def VelocityForDirection(self, dir):
        return self._velocity[dir]

class Character(sprite.Sprite):
    def __init__(self, img_name, **kwargs):
        self._whole_image = image.load(data.filepath(img_name))
        self._frame = 0
        self._direction = Direction.Stop
        self.flipped = False
        self._bb = AARectShape(Vector2(16, 8), 8, 8)
        self._inventory = {}
        self.blocked = False
        super(Character, self).__init__(self._cur_image())
    
    def _get_frame(self):
        assert 0 <= self._frame < _FRAME_COUNT, "%d is not a valid frame number" % self._frame
        return self._frame
        
    def _set_frame(self, i):
        if 0 <= i < _FRAME_COUNT:
            self._frame = i
            self.image = self._cur_image()
            
    frame = property(_get_frame, _set_frame)    
    
    def _set_direction(self, dir):
        assert type(dir)==int and (0 <= dir <= 21), "Bad direction %d" % dir
        self._direction = dir
        if dir in (Direction.West, Direction.NorthWest, Direction.SouthWest):
            self.flipped = True
        else:
            self.flipped = False
        self.frame = self.frame

    def _get_direction(self):
        if self.blocked:
            self._direction = Direction.Stop
        return self._direction
    direction = property(_get_direction, _set_direction)
    
    @property
    def velocity(self):
        return Direction().VelocityForDirection(self.direction)
        
    def BoundingBox(self):
        p = self.position
        self._bb.center = Vector2(p[0], p[1]-24)
        return self._bb
    
    cshape = property(BoundingBox)
        
    def _cur_image(self):
        img = self._whole_image.get_region(*_REGION[self.frame])
        if self.flipped:
            img = img.get_texture().get_transform(flip_x=True)
        return img
    
    def collidesWith(self, tile):
        def rectFromAABB(cshape):
            return Rect(cshape.center[0] - cshape.rx, cshape.center[1] - cshape.ry, cshape.rx*2, cshape.ry*2)
        hero_rect = rectFromAABB(self.cshape)
        tile_rect = rectFromAABB(tile)
        diff = hero_rect.intersect(tile_rect)
        x, y = self.position
        buffer = 0
        
        if self.direction in (Direction.North, Direction.NorthEast, Direction.NorthWest):
            y -= diff.height + buffer
        elif self.direction in (Direction.South, Direction.SouthEast, Direction.SouthWest):
            y += diff.height + buffer
        if self.direction in (Direction.East, Direction.NorthEast, Direction.SouthEast):
            x -= diff.width + buffer
        elif self.direction in (Direction.West, Direction.NorthWest, Direction.SouthWest):
            x += diff.width + buffer
        if self.direction == Direction.Stop:
            if diff.width < diff.height:
                x += (1 if x > diff.x else -1) * (diff.width + buffer)
            else:
                y += (1 if y > diff.y else -1) * (diff.height + buffer)
            
        self.position = x,y
        self.blocked = True
        self.parent.stopWalking()
        self.direction = Direction.Stop
    
    def inventory(self):
        return self._inventory.values()
    
    def aquireItem(self, item):
        self._inventory[item.name] = item
    
    def removeItem(self, item):
        return self._inventory.pop(item.name)
    
        
        
