#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Game main module.

Contains the entry point used by the run_game.py script.
'''
from __future__ import division
import cocos
from cocos.director import director
from cocos.scenes.transitions import FadeTransition
from cocos.actions.interval_actions import Delay, ScaleBy
from cocos.actions.instant_actions import CallFunc
from pyglet import font, window, event, image
import data
from character import Character, Walk, Blink, Direction
import tiled2cocos
from ConfigParser import SafeConfigParser
from string import upper,lower

reso_x, reso_y = 800, 576
n_scenes = 11

import threading
from Queue import Queue

from story import story
import items
import mission
from sound import sound

_key_values = {}
for v, n in window.key._key_names.iteritems():
    _key_values[n] = v

def symbol_value(value):
    '''Returns the key code for the string provided'''
    return _key_values.get(value, str(value))

from ui import UI, Message, plainLabel, addMessage

class KeyBindings(object):
    def __init__(self, forward, backward, left, right, action, inventory):
        self.set(forward, backward, left, right, action, inventory)
        
    def set(self, forward, backward, left, right, action, inventory):
        def sv(v): return symbol_value(upper(v))
        self.forward = sv(forward)
        self.backward = sv(backward)
        self.left = sv(left)
        self.right = sv(right)
        self.action = sv(action)
        self.inventory = sv(inventory)

    def _str_for(self, key_name):
        return window.key.symbol_string(getattr(self, key_name))
    forward_str = lambda self: self._str_for('forward')
    backward_str = lambda self: self._str_for('backward')
    left_str = lambda self: self._str_for('left')
    right_str = lambda self: self._str_for('right')
    action_str = lambda self: self._str_for('action')
    inventory_str = lambda self: self._str_for('inventory')
        
kb = KeyBindings('w', 's', 'a', 'd', 'space', 'i')
        
class GameLayer(cocos.layer.Layer):
    is_event_handler = True
    def __init__(self, hero, *args, **kwargs):
        super(GameLayer, self).__init__(*args, **kwargs)
        
        self.hero = hero
        self.hero.position = 400, 300
        self.add(self.hero)
        
        self._walk = None
        
        self.actors = [self.hero]
        
        self.active_key = set()
        self._dir = Direction()
        self._key_dir = {
            kb.forward: self._dir.DirFromName('north'),
            kb.backward: self._dir.DirFromName('south'),
            kb.left: self._dir.DirFromName('west'),
            kb.right: self._dir.DirFromName('east'),
        }

    def on_enter(self):
        super(GameLayer, self).on_enter()
        self.hero.parent = self
    
    def walk(self, dir):
        self.hero.direction = self._dir.CombineDirs(dir, self.hero.direction)
        if ((self._walk is None) or self._walk.done()) and not self.hero.blocked:
            print "  walk action"
            self._walk = self.hero.do(Walk())

    def stopWalking(self):
        print "WALK: stopped walking (moving)"
        if self.hero.are_actions_running() and self._walk is not None and not self._walk.done():
            self.hero.direction = Direction.Stop
            self.hero.remove_action(self._walk)
        self._walk = None
    
    def stopWalk(self, dir):
        self.hero.direction = self._dir.SubtractDirs(self.hero.direction, dir)
        if (self.hero.direction == self._dir.Stop):
            self.stopWalking()
    
    def on_key_press(self, key, modifiers):            
        print "Pressed: " + window.key.symbol_string(key)
        if key in self.active_key:
            print "   ***"+ window.key.symbol_string(key)+ " was already active***"
            return
        if key in (kb.forward, kb.backward, kb.left, kb.right):
            dir = self._key_dir[key]
            self.active_key.add(key)
            self.walk(dir)
            
        if key == kb.action:
            self.parent._doAction()

        if key == kb.inventory:
            self.parent.openInventory()
    
    def on_key_release(self, key, modifiers):
        def active_key(k, dir):
            return (k == dir) and (key in self.active_key)
        if active_key(key, kb.forward) \
        or active_key(key, kb.backward) \
        or active_key(key, kb.left) \
        or active_key(key, kb.right):
            print "Released: " + window.key.symbol_string(key)
            self.active_key.remove(key)
            self.stopWalk(self._key_dir[key])

class InventoryLayer(cocos.layer.Layer):
    is_event_handler=True
    def __init__(self, hero):
        super(InventoryLayer,self).__init__()
        self._icoords = [147, 320, 493]
        self._jcoords = [429, 279, 129]
        #self._missions_coords = 
        self._hero = hero
        background = cocos.sprite.Sprite(image.load(data.filepath('inventory.png')), position=(reso_x/2, reso_y/2))
        self._cur_inv = []
        self.add(background)
        self.add(plainLabel('MISSIONS', (612, reso_y - 65)), z=5)

    def _coords(self, i, j): # coordinates for cell i,j
        assert (i < 3) and (j < 3), "Invalid Cell"
        return self._icoords[i], self._jcoords[j]

    def _cell_coords(self, i):
        return self._coords(i % 3, i // 3)

    def on_enter(self):
        super(InventoryLayer,self).on_enter()
        for i in self._cur_inv:
            self.remove(i)
        self._cur_inv = []
        for i, inv in enumerate(self._hero.inventory()):
            nsprite = cocos.sprite.Sprite(inv.sprite.image, position=self._cell_coords(i), scale=4)
            self.add(nsprite)
            self._cur_inv.append(nsprite)
        y = reso_y - 110
        for i, mission in enumerate(story.getActiveMissions()):
            if mission.hidden: continue
            pl = plainLabel(mission.descr, (612, y-(i*25)), width=156, multiline=True)
            self.add(pl, z=5)
            self._cur_inv.append(pl)

    def on_key_press(self, key, modifiers):            
        print "Pressed: " + window.key.symbol_string(key)
        if key in (kb.inventory, window.key.ESCAPE):
            self.parent.closeInventory()        
        return event.EVENT_HANDLED
    
class WazScene(cocos.scene.Scene):
    def __init__(self, map, scene_index, hero):
        self._curMap = map
        self._mapIndex = scene_index
        sm = cocos.layer.ScrollingManager()
        sm.add(self._curMap)
        sm.set_focus(400,300)
        self.gameLayer = GameLayer(hero)
        self.invLayer = InventoryLayer(hero)
        self._doOnEnter = []
        
        self.ui = UI()
        
        self.entities = [] # objects to be interacted with in game
        self.actionable_entities = set()
        self.entityLayer = cocos.layer.Layer()
        
        super(WazScene, self).__init__(sm, self.gameLayer, self.entityLayer, self.ui)
                
        self._screensize = director.window.get_size()
        
        self.schedule(self.step)

    def setFirstThings(self, doOnEnter):
        self._doOnEnter = doOnEnter
        return self

    @property
    def hero(self):
        return self.gameLayer.hero
    
    def on_enter(self):
        print "Entering scene %d" % self._mapIndex
        super(WazScene, self).on_enter()
        for i in self._doOnEnter:
            i()
        self._doOnEnter = []
        self._refresh_entities()
    
    def on_exit(self):
        super(WazScene, self).on_exit()
        self.gameLayer.stopWalking()
        self.gameLayer.active_key.clear()
        
    def step(self, dt):
        # am I still on the map?
        story.step(dt)
        x, y = self.gameLayer.hero.position
        sx, sy = self._screensize
        if (x < 0):
            Waz.travel(Direction.West)
        if (x > sx):
            Waz.travel(Direction.East)
        if (y < 0):
            Waz.travel(Direction.South)
        if (y > sy):
            Waz.travel(Direction.North)
        self._collide()

    def openInventory(self):
        self.gameLayer.pause()
        self.ui.pause()
        self.entityLayer.pause()
        for i in self.entities:
            i.sprite.pause()
        self.add(self.invLayer, z=5)
    
    def closeInventory(self):
        self.remove(self.invLayer)
        self.entityLayer.resume()
        self.ui.resume()
        self.gameLayer.resume()
        for i in self.entities:
            i.sprite.resume()
        
    def _collide(self):
        self.gameLayer.hero.blocked = False
        hs = self.gameLayer.hero.cshape
        for i in self._curMap.doors:
            if i.cshape.overlaps(hs):
                Waz.enterDoor()
                return
        
        cw = None
        distance = 64
        for i in self._curMap.collidables:
            if i.cshape.overlaps(hs):
                d = i.cshape.distance(hs)
                if distance > d:
                    distance = d
                    cw = hs
        if cw is not None:
            self.gameLayer.hero.collidesWith(cw)
            
        old_entities = self.actionable_entities
        self.actionable_entities = set()        
        for i in self.entities:
            if i.cshape.distance(hs) < 32:
                if i not in old_entities:
                    i.playerNear()
                self.actionable_entities.add(i)
        for i in old_entities - self.actionable_entities:
            i.playerFar()
            
    def _refresh_entities(self):
        self.entities = []
        for i in story.getActiveMissions():
            self.entities.extend(i.entities(self._mapIndex))
        
        self.remove(self.entityLayer)
        self.entityLayer = cocos.layer.Layer()
        self.add(self.entityLayer, z=1)
        for i in self.entities:
            if i.visible:
                self.entityLayer.add(i.sprite)
                
    def _doAction(self):
        if len(self.actionable_entities) == 1:
            self.actionable_entities.pop().action(self.gameLayer.hero)

    def completeCurrentMission(self):
        m = [ i for i in story.getActiveMissions() if not i.hidden ]
        if len(m) > 0:
            m[0].end()
            self._refresh_entities()
            
class WazManager(event.EventDispatcher):
    def __init__(self):
        self._scenes = [None for i in xrange(n_scenes)]
        self._curMap = 1
        self._background_load = False
        self._maps = {}
        self._screensize = None
        self._hero = None
        self._music = sound.playMusic(data.filepath('BicyclingForDimSum.mp3'))
    
    def _load_map(self, index):
        name = "w-%d.tmx" % index
        if self._maps.has_key(name):
            return self._maps[name]
        l = tiled2cocos.load_map(data.filepath(name))
        self._maps[name] = l
        return l

    def loadScenes(self, *scenes):
        self._screensize = director.window.get_size()
        for i in scenes:
            self.scene(i)
        mission.createMissions()
        for m in story.listMissions():
            m.push_handlers(self)
    
    def scene(self, sceneIndex):
        if self._hero == None:
            self._hero = Character('hero.png')
        if self._scenes[sceneIndex] is None:
            print "Loading scene %i from disk" % sceneIndex
            self._scenes[sceneIndex] = WazScene(self._load_map(sceneIndex), sceneIndex, self._hero)
        return self._scenes[sceneIndex]

    def _switchToMap(self, index):
        self._curMap = index
        s = self.scene(index)
        director.replace(s)
        self.dispatch_event('on_switch_screen', index)
        return s
    
    def on_mission_complete(self, mission):
        if mission.next != '':
            if type(mission.next) == str:
                story.startMission(mission.next)
            else:
                for i in mission.next:
                    story.startMission(i)

    def on_mission_start(self, mission):
        if mission.notify_start:
            addMessage('Recieved mission: ' + mission.descr, 2)
            
    def travel(self, direction):
        assert self._curMap < 6, "Cannot travel from within a building"
        dist = 32
        self._screensize = director.window.get_size()
        if direction == Direction.West:
            assert self._curMap not in (2,5), "Cannot travel west from here (%d)" % self._curMap
            new_map = self._switchToMap(self._curMap+1)
            new_map.hero.position = self._screensize[0]-dist, new_map.hero.position[1]
        elif direction == Direction.East:
            assert self._curMap not in (0, 3), "Cannot travel east from here (%d)" % self._curMap
            new_map = self._switchToMap(self._curMap-1)
            new_map.hero.position = dist, new_map.hero.position[1]
        elif direction == Direction.North:
            assert self._curMap in (0, 2), "Cannot travel north from here (%d)" % self._curMap
            new_map = self._switchToMap(self._curMap + 3)
            new_map.hero.position = new_map.hero.position[0], dist
        elif direction == Direction.South:
            assert self._curMap in (3, 5), "Cannot travel south from here (%d)" % self._curMap
            new_map = self._switchToMap(self._curMap - 3)
            new_map.hero.position = new_map.hero.position[0], self._screensize[1]-dist

    def enterDoor(self):
        assert self._curMap != 0, "No Doors on Map 0"
        if self._curMap < 6:
            self._enterBuilding()
        else:
            self._exitBuilding()
            
    def _enterBuilding(self):
        new_map = self._switchToMap(self._curMap+5)
        if self._curMap == 7:
            new_map.hero.position = (544, 288)
        else:
            new_map.hero.position = (384, 192)

    def _exitBuilding(self):
        new_map = self._switchToMap(self._curMap - 5)
        if self._curMap == 2:
            new_map.hero.position = 96, 256
        else:
            new_map.hero.position = 384, 288

    def finale(self):
        director.replace(cocos.scene.Scene(EndLayer()))

WazManager.register_event_type('on_switch_screen')

Waz = WazManager()

class EndLayer(cocos.layer.Layer):
    is_event_handler = True
    text = (
        ( 'Well done! You visit your friend FRANCK and are cured of the curse.', 3),
        ( 'You have the success. Congratumalations.', 2),
        ( 'We couldn\'t have done it without you.', 2),
        )
    i = 0
    _font_size = 40
    def on_enter(self):
        super(EndLayer, self).on_enter()
        self._action = None
        self.label = cocos.text.Label(
            text='',
            font_name='Philosopher',
            font_size=self._font_size,
            x=400,
            y=288,
            anchor_x='center',
            anchor_y='center',
            width=600,
            multiline=True,
            color=(255,255,255,255))
        cont_label = cocos.text.Label(
            text='<%s> to skip' % kb.action_str(),
            font_name='Philosopher',
            font_size=10,
            x=700,
            y=20)
        self.add(self.label)
        self.add(cont_label)
        self.next()

    def end(self):
        import sys; sys.exit(0)

    def next(self):
        if self.i >= len(self.text):
            self.end()
            return
        if self._action is not None and not self._action.done(): self.remove_action(self._action)
        t = self.text[self.i][0]
        self.label.element.text = t
        if len(self.text[self.i]) == 3: 
            self.label.element.set_style('font_size', self.text[self.i][2])
        else:
            self.label.element.set_style('font_size', self._font_size)
        self._action = self.do(Delay(self.text[self.i][1]) + CallFunc(self.next))
        self.i += 1

    def on_key_press(self, key, modifiers):
        if key == kb.action:
            self.next()

class StoryLayer(cocos.layer.Layer):
    is_event_handler = True
    text = (
        ('MUTATE: Verb /mju:\'teit/\nChange or cause to change in form or nature.', 4),
        ('You fathers:\nif your children ask for a fish,\ndo you give them a snake instead?\n'+
         'Or if they ask for an egg,\ndo you give them a scorpion?\nOf course not!\n\nLuke 11', 8, 30),
        #('Important Hint:\nA Monkey Flings Poo', 2),
        ('Entering the bazaar from the south, you feel pangs of hunger.', 3),
        ('Looking around for something to eat, you see an APPLE to the EAST.', 3),
        ('As you head that direction, you stumble over a dirty looking man named Martin.', 3),
        ('He begins shouting.', 0.1),)
    i = 0
    _font_size = 40
    def on_enter(self):
        super(StoryLayer, self).on_enter()
        self._action = None
        self.label = cocos.text.Label(
            text='',
            font_name='Philosopher',
            font_size=self._font_size,
            x=400,
            y=288,
            anchor_x='center',
            anchor_y='center',
            width=600,
            multiline=True,
            color=(255,255,255,255))
        cont_label = cocos.text.Label(
            text='<%s> to skip' % kb.action_str(),
            font_name='Philosopher',
            font_size=10,
            x=700,
            y=20)
        self.add(self.label)
        self.add(cont_label)
        self.next()

    def end(self):
        director.replace(Waz.scene(1).setFirstThings([lambda: story.startMission('people'), lambda: story.startMission('GetFruit')]))

    def next(self):
        if self.i >= len(self.text):
            self.end()
            return
        if self._action is not None and not self._action.done(): self.remove_action(self._action)
        t = self.text[self.i][0]
        self.label.element.text = t
        if len(self.text[self.i]) == 3: 
            self.label.element.set_style('font_size', self.text[self.i][2])
        else:
            self.label.element.set_style('font_size', self._font_size)
        self._action = self.do(Delay(self.text[self.i][1]) + CallFunc(self.next))
        self.i += 1

    def on_key_press(self, key, modifiers):
        if key == kb.action:
            self.next()

_dialog_tree = {
    'kindclarissa': [
        Message('Your apple turned into a monkey? Oh Dear.', 2),
        Message('Yes, I\'ve seen this before. Martin loves his curses.', 2),
        Message('This curse is called Mutate.', 2),
        Message('It changes whatever you pick up into something else.', 2),
        Message('Go EAST and see FRANCK, he\'ll know what to do.', 2),
        ],
    'wisefranck': [
        Message('Martin is such a joker, I could see you coming from at least 5 meters.', 3),
        Message('I can make you the antidote, but I need two items.', 3),
        Message('Bring me:', 1),
        Message('a recently used TOILET BRUSH', 2),
        Message('and the BOOK OF POTIONS', 2),
        Message('Now would be a good time to mention you can see your inventory with <%s>' % lower(kb.inventory_str()), 4),
        Message('Well, get to it!', 2)
        ],
    'henry': [
        Message('Huh? You want our nasty old TOILET BRUSH?', 2),
        Message('That\'s really gross buddy.', 2)
        ],
    'henry2': [
        Message('Tell you what? You can have it on one condition', 2),
        Message('I strongly dislike JOHN.', 1),
        Message('If you humiliate him, we can talk', 2),
        Message('Find him south of here', 2)
        ],
    'henry3': [
        Message('Well? Get to it!', 1),
        ],
    'henry4': [
        Message('Monkey Poo? Oh man.', 1),
        Message('I thought that TOILET BRUSH was nasty.', 2),
        Message('PHILIP, throw it over!', 2),
        Message('Here you go. Nasty.', 4),
        ],
    'JOHN': [
        Message('Hello! Have an apple.', 1),
        ],
    'JOHN1': [
        Message('Hello again! I\'m out of apples.', 1),
        Message('You took my last one.', 1),
        ],
    }

def dialogTree(name):
    return _dialog_tree[name]

def main():
    cp = SafeConfigParser()
    cp.read('settings.ini')
    fullscreen = cp.getint('main', 'fullscreen') == 1
    
    cocos.director.director.init(width=int(reso_x), height=int(reso_y), caption="Wazemmes", fullscreen=fullscreen)
    
    kb.set(
        cp.get('keys', 'forward'),
        cp.get('keys', 'backward'),
        cp.get('keys', 'left'),
        cp.get('keys', 'right'),
        cp.get('keys', 'action'),
        cp.get('keys', 'inventory'))

    font.add_file(data.filepath('megrim/Megrim.ttf'))
    font.add_file(data.filepath('philosopher/Philosopher-Regular.ttf'))
    font.add_file(data.filepath('tangerine/Tangerine_Regular.ttf'))
    font.add_file(data.filepath('unifrakturmaguntia/UnifrakturMaguntia-Book.ttf'))

    Waz.loadScenes(*[i for i in xrange(n_scenes)])
    cocos.director.director.run(cocos.scene.Scene(StoryLayer()))
