# UI Layer

from cocos.director import director
from cocos.layer import Layer
from cocos.text import Label
from cocos import actions
from cocos.actions.instant_actions import CallFunc
from cocos.actions.interval_actions import Delay
from pyglet import image
from pyglet import gl
from pyglet import graphics
import data

class Message(object):
    def __init__(self, text, display_time, display_in=0, font_name='Philosopher', font_size=15):
        self.text = text
        self.display_time = display_time
        self.wait_seconds = display_in
        self.font_name = font_name
        self.font_size = font_size

_queue = []
def addMessage(text, display_time, display_in=0, font_name='Philosopher', font_size=15):
    _queue.append(Message(text, display_time, display_in, font_name, font_size))

class BlinkDialogEvent(actions.Action):
    def init(self, on_time, off_time):
        self._on_time = on_time
        self._off_time = off_time
        self._initial = True
        self._elapsed = 0.0
    
    @property
    def _duration(self):
        return self._on_time if self.target.visible else self._off_time
        
    def step(self, dt):
        if self._initial:
            self._initial = False
            return
        self._elapsed += dt
        if self._elapsed >= self._duration:
            self._elapsed = 0.0
            self.target.visible = not self.target.visible
    
    def stop(self):
        self.target.visible = True

def plainLabel(text, position, **kwargs):
    return Label(text,
                 position,
                 font_name='Philosopher',
                 font_size=15,
                 color=(0,0,0,255),
                 anchor_x='left',
                 anchor_y='top',
                 **kwargs)
    
class DialogBox(Label):
    def __init__(
        self, 
        text='', 
        position=(0, 0), 
        font_name='Philosopher',
        font_size=15,
        left_img='leftQuote.png', 
        right_img='rightQuote.png'):
        super(DialogBox, self).__init__(
            text=' '+text, 
            position=(position[0]+7, position[1]), 
            font_name=font_name,
            font_size=font_size,
            anchor_x='left', 
            anchor_y='top', 
            color=(255,255,255,255))
        self._left = image.load(data.filepath(left_img)).get_texture()
        self._right = image.load(data.filepath(right_img)).get_texture()
        self.element.set_style('background_color', (0, 0, 0, 255))
    
    def draw(self):
        gl.glPushMatrix()
        self.transform()
        self.element.draw()
        cw, ch = self.element.content_width, self.element.content_height
        gl.glPushMatrix()
        gl.glTranslatef(-7, -ch, 0)
        self._left.blit(0,0,0)
        gl.glPopMatrix()
        gl.glTranslatef(cw, -ch, 0)
        self._right.blit(0,0,0)
        gl.glPopMatrix()

class UI(Layer):
    def __init__(self, *args, **kwargs):
        super(UI, self).__init__(*args, **kwargs)
        self._action = self.do(Delay(0))
        self.schedule(self.step)
        self._dialog_list = {}

    def _makeDialogBox(self, item, text):
        return DialogBox(text, 
            (item.sprite.position[0]+(item.sprite.width/2), item.sprite.position[1]+(item.sprite.height/4)))
        
    _dialogItem = {}
    def dialogBox(self, item, text):
        db = self._makeDialogBox(item, text)
        self.add(db)
        self._dialogItem[item] = db
        return db

    def dialogList(self, item, message_list):
        if len(message_list) == 0: return
        if self._dialog_list.has_key(item):
            l = self._dialog_list[item]
        else:
            l = []
            self._dialog_list[item] = l
        for i in message_list:
            db = self._makeDialogBox(item, i.text)
            db.visible = False
            self.add(db)
            l.append([i.display_time, db])
        l[0][1].visible = True
    
    def step(self, dt):
        if self._action.done() and (len(_queue) > 0):
            msg = _queue.pop(0)
            self._action = self.do(Delay(msg.wait_seconds) +
                                   (CallFunc(lambda: self.infoBox(
                                       msg.text,
                                       display_for=msg.display_time,
                                       font_name=msg.font_name,
                                       font_size=msg.font_size)) |
                                   Delay(msg.display_time)))
        for k,v in [(k,v) for k, v in self._dialog_list.iteritems()]:
            if len(v) > 0:
                m = v[0]
                m[0] -= dt
                if m[0] <= 0:
                    m[1].visible = False
                    self.remove(m[1])
                    v.pop(0)
                    if len(v) > 0:
                        m = v[0]
                        m[1].visible = True
                    else:
                        k.doneTalking()
                    
    def removeDialogFromItem(self, item):
        if not self._dialogItem.has_key(item): return
        db = self._dialogItem[item]
        self.remove(db)
        db = None
        del self._dialogItem[item]
            
    def infoBox(self, text, position=(400, 20), display_for=0, font_name='Philosopher', font_size=15):
        db = DialogBox(text, position, left_img='leftEdge.png')
        db.position = (800 - db.element.content_width) / 2, db.position[1]
        self.add(db)
        if display_for > 0:
            self.do(Delay(display_for) + CallFunc(lambda: self.removeInfoBox(db)))
        return db
    
    def removeInfoBox(self, infoBox):
        self.remove(infoBox)

def getUI():
    return director.scene.ui
